"""
```
ROCCO: [R]obust [O]pen [C]hromatin Dection via [C]onvex [O]ptimization.

PyPI : https://pypi.org/project/rocco/

GitHub: https://github.com/nolan-h-hamilton/ROCCO/

Demo: https://github.com/nolan-h-hamilton/ROCCO/blob/main/demo/demo.ipynb

Documentation: https://nolan-h-hamilton.github.io/ROCCO/

Paper: https://doi.org/10.1101/2023.05.24.542132


usage: rocco [-h] {gwide,chrom,prep,budgets,get_sizes} ...
  {gwide,chrom,prep,budgets,get_sizes}
    gwide               run rocco genome-wide/on multiple chromosomes (ROCCO_gwide.py)
    chrom               run ROCCO on a single chromosome (ROCCO_chrom.py)
    prep                Preprocess BAM files (prep_bams.py)
    budgets             Compute a budget (maximum fraction of basepairs that can be selected as 'open') for
                        each chromosome ranked by average read-density observed in the supplied BAM files (est_budgets.py)
    get_sizes           download sizes file for a genome in the ucsc genome registry
options:
  -h, --help            show this help message and exit

note: 
rocco gwide --> rocco/ROCCO_gwide.py
rocco chrom --> rocco/ROCCO_chrom.py
rocco prep --> rocco/prep_bams.py
rocco budgets --> rocco/est_budgets.py
```
"""
#!/usr/bin/env python
import os
import argparse
from . import rocco_aux
from . import ROCCO_chrom
from . import ROCCO_gwide
from . import prep_bams
from . import est_budgets
from . import locus
from . import loci


def subcommand_chrom(args):
    ROCCO_chrom.main(args)
    return True

def subcommand_gwide(args):
    ROCCO_gwide.main(args)
    return True

def subcommand_prep(args):
    prep_bams.main(args)
    return True

def subcommand_budgets(args):
    est_budgets.main(args)
    return True


def main():
    parser = argparse.ArgumentParser(description="ROCCO: [R]obust [O]pen [C]hromatin Dection via [C]onvex [O]ptimization.\n\nPyPI : https://pypi.org/project/rocco/ \nGitHub: https://github.com/nolan-h-hamilton/ROCCO/ \nDemo: https://github.com/nolan-h-hamilton/ROCCO/blob/main/demo/demo.ipynb \nPaper: https://doi.org/10.1101/2023.05.24.542132\n", add_help=True, formatter_class=argparse.RawTextHelpFormatter)
    subparsers = parser.add_subparsers(dest="command")

    # 'gwide' subcommand parameters
    parser_subcommand_gwide = subparsers.add_parser("gwide", help='run rocco genome-wide/on multiple chromosomes (ROCCO_gwide.py)')
    parser_subcommand_gwide.add_argument('-p', '--param_file', required=True,
                                         help='A CSV file specifying the parameters for each chromosome. See: https://github.com/nolan-h-hamilton/ROCCO/blob/main/demo/demo_params.csv')
    parser_subcommand_gwide.add_argument('-b', '--budget', type=float, default=.035)
    parser_subcommand_gwide.add_argument('-g', '--gamma', type=float, default=1.0)
    parser_subcommand_gwide.add_argument('-t', '--tau', type=float, default=0.0)
    parser_subcommand_gwide.add_argument('--c1', type=float, default=1.0)
    parser_subcommand_gwide.add_argument('--c2', type=float, default=1.0)
    parser_subcommand_gwide.add_argument('--c3', type=float, default=1.0)
    parser_subcommand_gwide.add_argument('-N', '--rr_iter', type=int, default=50)
    parser_subcommand_gwide.add_argument('--solver', type=str, default="ECOS")
    parser_subcommand_gwide.add_argument('--bed_format', type=int, default=3)
    parser_subcommand_gwide.add_argument('--identifiers', default=None)
    parser_subcommand_gwide.add_argument('--outdir', default='.')
    parser_subcommand_gwide.add_argument('--combine', default=None)
    parser_subcommand_gwide.add_argument('--multi', default=False, action='store_true')
    parser_subcommand_gwide.add_argument('--verbose', default=False, action="store_true")

    # 'chrom' subcommand parameters
    parser_subcommand_chrom = subparsers.add_parser("chrom", help='run ROCCO on a single chromosome (ROCCO_chrom.py)')
    parser_subcommand_chrom.add_argument('--start', type=int, default=-1)
    parser_subcommand_chrom.add_argument('--end', type=int, default=-1)
    parser_subcommand_chrom.add_argument('--locus_size', type=int, default=-1)
    parser_subcommand_chrom.add_argument('-b', '--budget', type=float, default=.035)
    parser_subcommand_chrom.add_argument('--chrom', type=str, help='')
    parser_subcommand_chrom.add_argument('--wig_path', type=str, default=os.getcwd())
    parser_subcommand_chrom.add_argument('-g', '--gamma', type=float, default=1.0)
    parser_subcommand_chrom.add_argument('-t', '--tau', type=float, default=0.0)
    parser_subcommand_chrom.add_argument('--c1', type=float, default=1.0)
    parser_subcommand_chrom.add_argument('--c2', type=float, default=1.0)
    parser_subcommand_chrom.add_argument('--c3', type=float, default=1.0)
    parser_subcommand_chrom.add_argument('--solver', type=str,default="ECOS")
    parser_subcommand_chrom.add_argument('--bed_format', type=int, default=6)
    parser_subcommand_chrom.add_argument('--identifiers', default=None)
    parser_subcommand_chrom.add_argument('--outdir', type=str, default='.')
    parser_subcommand_chrom.add_argument('-N', '--rr_iter', type=int, default=50)
    parser_subcommand_chrom.add_argument('--verbose', default=False, action="store_true")

    # 'prep' subcommand parameters
    parser_subcommand_prep = subparsers.add_parser("prep", help='Preprocess BAM files (prep_bams.py)')
    parser_subcommand_prep.add_argument('-i', '--bamdir', default='.', type=str)
    parser_subcommand_prep.add_argument('-o', '--outdir', default='.')
    parser_subcommand_prep.add_argument('-s', '--sizes', default='hg38')
    parser_subcommand_prep.add_argument('-L', '--interval_length', default=50)
    parser_subcommand_prep.add_argument('-c', '--cores', type=int, default=1)
    parser_subcommand_prep.add_argument('--multi', default=True)
    parser_subcommand_prep.add_argument('--bstw_path', default=None, help='deprecated')

    # 'budgets' subcommand parameters
    parser_subcommand_budgets = subparsers.add_parser("budgets", help='Compute a budget for each chromosome ranked by average read-density observed in the supplied BAM files (est_budgets.py)')
    parser_subcommand_budgets.add_argument('-i', '--bamdir', type=str)
    parser_subcommand_budgets.add_argument('-s', '--sizes', type=str)
    parser_subcommand_budgets.add_argument('-a', type=float, default=0.0)
    parser_subcommand_budgets.add_argument('-b', type=float, default=0.05)
    parser_subcommand_budgets.add_argument('--desired_avg', type=float, default=-1.0)
    parser_subcommand_budgets.add_argument('--index', default=False, action='store_true')

    # 'get_sizes' subcommand parameters
    parser_subcommand_get_sizes = subparsers.add_parser("get_sizes", help="download sizes file for a genome in the ucsc genome registry")
    parser_subcommand_get_sizes.add_argument('-g', '--genome', type=str, default='hg38')

    args = vars(parser.parse_args())
    if args['command'] == "gwide":
        subcommand_gwide(args)
    elif args['command'] == "chrom":
        subcommand_chrom(args)
    elif args['command'] == 'prep':
        subcommand_prep(args)
    elif args['command'] == 'budgets':
        subcommand_budgets(args)
    elif args['command'] == 'get_sizes':
        rocco_aux.get_size_file(args['genome'])
    else:
        parser.print_help()

if __name__ == "__main__":
    main()



r"""
Loci object: ROCCO creates a Locus object (see `locus.py`) for each
$\ell_i, \forall i = 1 \ldots n$. These Locus objects are then linked
together as a Loci object, defined in this code.

The necessary functionality to construct the signal matrix,
$\mathbf{S}_{chr}$, is implemented by way of the Loci object.
"""
import copy
import math
import random
import warnings
import numpy as np
import cvxpy as cp
from scipy import stats
from .locus import Locus

class LociIterator:
    """
    Iterator for traversing the Loci object

    Args:
        head (Locus): first Locus object in the Loci object

    Returns:
        LociIterator: an iterator object

    Raises:
        StopIteration: raised if end of genomic region is reached during iteration
    """
    def __init__(self, head):
        self.current = head

    def __iter__(self):
        return self

    def __next__(self):
        if not self.current:
            raise StopIteration

        item = copy.copy(self.current)
        self.current = self.current.right
        return item

class Loci:

    def __init__(self, head=None, tail=None):
        self.head = head
        self.tail = tail

    def __iter__(self):
        return LociIterator(self.head)

    def is_empty(self) -> bool:
        """
        Check if the Loci object is empty

        Returns:
            bool: True if Loci object is empty--False otherwise.
        """
        if self.head is None and self.tail is None:
            return True
        return False

    def append_locus(self,locus):
        """
        Append a new locus as the tail of the Loci object.

        Args:
            locus (Locus): The locus to be appended.

        Returns:
            None
        """
        if  self.is_empty():
            self.head = locus
            self.tail = locus
            locus.left=None
            locus.right=None
        else:
            tail_cpy = self.tail
            self.tail = locus
            locus.right = None
            locus.left = tail_cpy
            tail_cpy.right=locus

    def combine(self,head_,tail_):
        """Combine contiguous accessible regions

        This function takes the starting and ending node
        for a sequence of contiguous loci (i.e., a sequence
        such that all loci satisfy `Locus.accessible = 1`) and
        merges them into a single locus object. It defines
        the new locus's signal vector as the sum
        of signal vectors among loci in the contiguous
        sequence.

        Args:
            head_ (Locus): first locus object in sequence of contiguous
              selected loci
            tail_ (Locus): last locus object in sequence of contiguous
              selected loci
        """
        cpy = head_
        accessible_ = 0
        size_ = 0
        sig_matrix = []
        subloc = Loci()

        accessible_= 0
        if head_.accessible == 1:
            accessible_ = 1

        while head_ is not None and head_.position <= tail_.position:
            size_ += head_.size
            sig_matrix.append(head_.sig_data)
            subloc.append_locus(copy.copy(head_))
            head_ = head_.right

        sig_matrix = np.array([np.array(sig) for sig in sig_matrix]).transpose()
        sig_data_ = np.sum(sig_matrix,axis=1)

        # create new locus
        new_loc = Locus(position=cpy.position,left=cpy.left,right=head_,
                        size=size_, parent=None, subloci=subloc,
                        sig_data=sig_data_,accessible=accessible_)

        if head_ is None:
            self.tail = new_loc
        else:
            head_.left = new_loc
            cpy.right = new_loc

        if cpy.left is not None:
            cpy.left.right = new_loc
        else:
            self.head = new_loc
        s_head_ = subloc.head
        while s_head_ is not None and s_head_.position <= tail_.position:
            s_head_.parent=new_loc
            s_head_ = s_head_.right

    def as_list(self) -> list:
        """
        Get a list representation of the Loci object.

        Returns:
            list: list representation of the Loci object, where the Locus elements
                are ordered as in the Loci object.
        """
        head_ = self.head
        list_repr = []
        while head_ is not None:
            list_repr.append(copy.copy(head_))
            head_ = head_.right
        return list_repr

    def combine_selected(self):
        r"""
        Find adjacent accessible regions and combine them into a single locus

        This function identifies sequences of locus objects, $\ell_i$, satisfying
        $\ell_i$`.accessible > 0` and combines them into a single locus via `Loci.combine()`.


        Notes:
            - `Loci.combine()` defines signal data for the created locus as the vector sum of
                signal data for each Locus object comprising the contiguous accessible region.

        Returns:
            None
        """
        head_ = self.head
        while head_ is not None:
            to_combine = []
            head_cpy = head_
            while head_cpy is not None and head_cpy.accessible > 0 :
                to_combine.append(head_cpy)
                head_cpy = head_cpy.right
            if len(to_combine) > 1:
                self.combine(to_combine[0], to_combine[-1])
            head_ = head_cpy
            if head_ is not None:
                head_ = head_.right

    def get_sig_mat(self) -> np.ndarray:
        r"""
        Builds the signal matrix $\mathbf{S}_{chr}$ as defined in the paper.

        Each Locus object's `sig_data` attribute is a $K \times 1$ column
        vector, so $\mathbf{S}_{chr}$ is a $K \times n$ matrix.

        Returns:
            None
        """
        sig_mat = []
        head_ = self.head
        while head_ is not None:
            sig_mat.append(head_.sig_data)
            head_ = head_.right
        return np.array(sig_mat)

    def g3_vec(self, metric_vec) -> np.ndarray:
        r"""
        Compute locus score term $g_3$ for each locus.

        Args:
            metric_vec (np.ndarray): (by default) a $1 \times n$ vector of
                median signal values at each locus.

        Returns:
            np.ndarray: A vector $g_3(i), \forall i=1 \ldots n$
        """
        v_vec = []
        for i,Loc in enumerate(metric_vec):
            if i == 0:
                vel = (abs(metric_vec[i] - metric_vec[i+1]))
            if i == len(metric_vec)-1:
                vel = abs(metric_vec[i] - metric_vec[i-1])
            else:
                vel = max(abs(metric_vec[i] - metric_vec[i+1]),
                      abs(metric_vec[i] - metric_vec[i-1]))

            vel /= metric_vec[i]+1
            v_vec.append(vel)
        return np.array(v_vec)

    def score_loci(self, tau: float = 0.0, c1: float = 1.0, c2: float = 1.0, c3: float = 1.0) -> np.ndarray:
        r"""
        compute locus score vector, i.e., $\mathcal{S}(i), \forall i = 1 \ldots n$
        given the $K \times n$ signal matrix $\mathbf{S}_{chr}$.

        Args:
            tau (float): median signal threshold (see $\tau$ in paper)
            c1 (float): weight for the first term in  $\mathcal{S}(i)$,
                (median signal) $g_1$
            c2 (float): weight for second term in $\mathcal{S}(i)$,
                (MAD dispersion) $g_2$
            c3 (float): weight for third term in  $\mathcal{S}(i)$, $g_3$

        Returns:
            np.ndarray: the $n$-dimensional score vector $\mathcal{S}(i), \forall i = 1 \ldots n$
        """
        sig_mat = self.get_sig_mat()
        med_vec = np.median(sig_mat,axis=1) # g_1
        mad_vec = stats.median_abs_deviation(sig_mat,axis=1) # g_2
        vel_vec = self.g3_vec(med_vec) # g_3
        s_vec = c1*med_vec - c2*mad_vec + c3*vel_vec

        for i, val in enumerate(s_vec):
            if med_vec[i] <= tau:
                s_vec[i] = 0
        return s_vec

    def run_rr(self, lp_sol, N, loci_scores, budget, gam, eps=1e-5) -> np.ndarray:
        r"""
        Carry out the $\texttt{RR}$ rounding procedure to return a
        'good' integral solution with reference to the LP solution

        Args:
            lp_sol (numpy.ndarray): The LP solution
            N (int): Number of RR solutions to evaluate. If $N \leq 0$, the
                $\texttt{floor\_eps}$ rounding protocol is applied in which
                any decision variable satisfying $\ell_i + \epsilon < 1$
                (`lp_sol[i] + eps < 1`) is rounded down to zero.
            loci_scores (list): mathcal{S} for each locus
            budget (float): budget parameter
            gam (float): gamma
            eps (float): epsilon value for "floor" rounding (Default: eps=1e-5)

    Returns:
        np.ndarray: the integral $\texttt{RR}$ solution, OR $\texttt{floor\_eps}$ sol if $N \leq 0$.
    """
        n = len(loci_scores)
        init_sol = np.floor(lp_sol + eps)
        init_score = (-loci_scores@init_sol
                       + gam*np.sum(np.abs(np.diff(init_sol,1))))
        if N <= 0:
            num_selected = np.sum(init_sol)
            if num_selected > math.floor(n*budget):
                # this shouldn't really occur for reasonable `eps`,
                # but a warning is included for completeness.
                warnings.warn('floor solution with eps={eps} exceeds budget by {num_selected - math.floor(n*budget)} selections')
            return init_sol

        rr_sol = init_sol
        best_score = init_score
        for j in range(N):
            ell_rand_n = np.zeros(n, dtype=np.int8)
            for i, ell_i in enumerate(lp_sol):
                if random.random() <= ell_i:
                    ell_rand_n[i] = 1
                else:
                    ell_rand_n[i] = 0

            score = (-loci_scores@ell_rand_n
                       + gam*np.sum(np.abs(np.diff(ell_rand_n,1))))
            is_feas = (np.sum(ell_rand_n) <= math.floor(n*budget))
            if is_feas and score < best_score:
                rr_sol = ell_rand_n
                best_score = score

        return rr_sol

    def rocco_lp(self, budget: float = .035, tau: float = 0, gam: float = 1, c1: float = 1,
             c2: float = 1, c3: float = 1, verbose_: bool = False, solver: str = "ECOS",
             N: int = 50) -> cp.Problem:
        r"""
        Solve relaxed problem as an LP and assign binary accessibility
        prediction to each locus using the $\texttt{RR}$ or $\texttt{floor\_eps}$
        procedure.

        Args:
            budget (float): budget constraint
            tau (float): tau in $\mathcal{S}(i)$
            gam (float): gamma
            c1 (float): c1 value in $\mathcal{S}(i)$
            c2 (float): c2 value in $\mathcal{S}(i)$
            c3 (float): c3 value in $\mathcal{S}(i)$
            N (int): RR procedure iterations. If N <= 0,
                \texttt{floor\_eps} procedure is applied...
                See `Loci.run_rr()`
            verbose_ (bool): Verbosity flag for the solver
            solver (str): Defaults to `ECOS`.

        Returns:
            cp.Problem: a CVXPY problem object
    """
        loci_scores = self.score_loci(tau, c1, c2, c3) # S_i, i=1...n
        n = len(loci_scores)
        # define problem in CVXPY
        ell = cp.Variable(n)
        z = cp.Variable(n-1)
        constraints = [ell <= 1,
                          ell >= 0,
                          cp.sum(ell) <= math.floor(budget*n),
                          z >= cp.diff(ell,1),
                          z >= -1*cp.diff(ell,1)]
        problem = cp.Problem(cp.Minimize(-loci_scores@ell + gam*cp.sum(z)),
                             constraints)

        # Refer to `Solve method options` at
        # https://www.cvxpy.org/tutorial/advanced/index.html
        if solver == "ECOS":
            problem.solve(solver=cp.ECOS, verbose=verbose_)

        if solver == "PDLP":
            try:
                problem.solve(solver=cp.PDLP, verbose=verbose_)
            except Exception as ex:
                print("Ensure PDLP solver is available.\
                    Can be installed via pip:\
                    'pip install ortools==9.3.10497'")

        if solver == "MOSEK":
            # https://docs.mosek.com/latest/pythonapi/parameters.html
            MOSEK_OPTS = {'MSK_IPAR_NUM_THREADS': 1,
                        'MSK_DPAR_INTPNT_TOL_REL_GAP': .005,
                        'MSK_IPAR_PRESOLVE_LINDEP_ABS_WORK_TRH':10,
                        'MSK_IPAR_INTPNT_SOLVE_FORM':2}
            try:
                problem.solve(cp.MOSEK, mosek_params=MOSEK_OPTS, verbose=verbose_)
            except Exception as ex:
                print("Ensure a valid MOSEK license file is\
                available in your home directory and that the\
                mosek-python interface is installed, e.g.,\
                via 'pip install mosek'")
                raise ex

        lp_sol = problem.variables()[0].value
        sol_rr  = self.run_rr(lp_sol, N, loci_scores, budget, gam)

        head_ = self.head
        i = 0
        while head_ is not None:
            head_.accessible = sol_rr[i]
            head_ = head_.right
            i += 1

        return problem



import sys
sys.path.append('../')
from locus import Locus
from loci import Loci
import numpy as np
import pandas as pd

def log(txt, script):
    print(f'{script}: {txt}')

signal_matrix = np.array([[0,0,0,0,0,2,10,2,0,0], [0,0,0,0,0,1,9,1,0,0],[0,0,0,0,0,3,11,3,0,0]])
signal_matrix = pd.DataFrame(signal_matrix)
signal_matrix.columns = [i*50 for i in range(0,10)]
loci_object = Loci()
size_ = 50
for i, loc in enumerate(signal_matrix.columns):
    new_loc = Locus(position=0 + i * size_,
                    size=size_,
                    sig_data=signal_matrix[loc])
    if i in [5,6,7]:
        new_loc.accessible = 1
    loci_object.append_locus(new_loc)
assert loci_object.is_empty() == False, "loci_object empty"
assert loci_object.head.position == 0, "incorrect position for first locus/first locus d.n.e"
assert loci_object.head.right.position == 50, "first locus points to wrong locus/first locus `right` d.n.e"
assert loci_object.head.right.left.position == 0, "locus left/right attributes not set correctly"


list_rep = loci_object.as_list()
assert len(list_rep) == 10, "list representation missing a locus"
for loc_obj in list_rep:
    assert len(loc_obj.sig_data) == 3, "missing locus signal data"


loci_object.combine_selected()
list_rep = loci_object.as_list()
i = 0
for loc_obj in loci_object:
    if i == 5:
        assert loc_obj.size == 150, "combined locus has incorrect size"
        i += 1
    else:
        assert loc_obj.size == 50, "combined locus has incorrect size"
        i += 1
assert i == 8, "combined loci object should have length 8"

log('all cases passed', sys.argv[0])


"""
Runs multiple `rocco chrom` jobs, which execute ROCCO  on each respective
chromosome. Chromosome-specific parameters are collected in a CSV file specified
with argument `-p --param_file`. Consider using the `budgets` subcommand
to set chromosome-specific budgets based on read densities.

If any parameter entry in the CSV file `--param_file` is set to `NULL`, the script uses the
corresponding default value specified by the command-line arguments.

Supports simultaneous execution of multiple `rocco chrom` jobs via `--multi`.

The output files generated by `rocco chrom` are stored in the specified output
directory (`--outdir`, default is current working directory). If the `combine`
argument is provided not `None`, then the script combines the chromosome-specific
bed files into a single output file with the specified name.

A note on `--solver`: The default solver is ECOS, which is free, open-source, and
conveniently installed alongside cvxpy. ECOS should prove sufficient in general, but
using a commerical (e.g., MOSEK) solver can improve speed of execution. *Free* academic
MOSEK licenses can be obtained *instantly* with a valid .edu email address: https://www.mosek.com/products/academic-licenses/
Generous free trials are also available for commercial use.

After obtaining the license (`mosek.lic`, or something similar), create a folder `mosek` in
your home directory and move the license file to it.

```
mkdir ~/mosek
mv mosek.lic ~/mosek
```

Then, run:

```
pip install mosek
```

After these steps, you should be able to specify MOSEK with
with `--solver MOSEK`


"""

import os
import sys
import argparse
import subprocess
import tempfile
from . import rocco_aux

def get_params(param_file: str, budget: float, gamma: float, tau: float,
               c1: float, c2: float, c3: float) -> list:
    """
    Grabs parameters for each chromosome from `param_file`

    This function collects parameters from each chromosome-row of
    `params.csv`. If a `NULL` entry is encountered, it is replaced
    by the CLI-specified default (see `args`).

    Args:
        param_file (str) : a CSV file with a row containing parameter vals\
            to use for each chromosome.

    Returns:
        list: a list of lists, with each element containing
          the necessary parameters to run a `rocco chrom` job.
    """
    defaults = [None, None, budget, gamma, tau, c1, c2, c3]
    chrom_params = []
    with open(param_file, mode='r', encoding='utf-8') as par_file:
        header = True
        for line in par_file:
            if header is True:
                header = False
                continue
            if ',' not in line:
                continue
            line = line.strip()
            line = line.split(',')
            for i, entry in enumerate(line):
                if entry == 'NULL':
                    # replace NULL entries with the genome-wide default
                    line[i] = defaults[i]
            chrom_params.append(line)
    return chrom_params


def call_rocco(chrom, wig_path, budget, gamma, tau, c1, c2, c3, solver,
               bed_format, verbose=False, N=50, identifiers=None,
               outdir='.') -> str:
    r"""
    Formats a command to run `rocco chrom` for a given chromosome

    Args:
        chrom (str): chromosome name. Example: `chr1`
        wig_path (str): path to the *directory* containing samples' signal
            tracks for `chrom`
        budget (float): budget constraint
        tau (float): tau in $\mathcal{S}(i)$
        gam (float): gamma parameter in $f(\mathbf{\ell})$
        c1 (float): c1 value in $\mathcal{S}(i)$
        c2 (float): c2 value in $\mathcal{S}(i)$
        c3 (float): c3 value in $\mathcal{S}(i)$
        N (int): RR procedure iterations. If N <= 0,
            $\texttt{floor\_eps}$ procedure is applied...
            See `Loci.run_rr()`
        verbose_ (bool): Verbosity flag for the solver
        solver (str): the solver to use--either "ECOS" or "MOSEK"
    Returns:
        str: a formatted rocco chrom command with the appropriate parameters.
        substituted.
    """
    cli_args = ["rocco", "chrom",
                '--chrom', chrom,
                '--wig_path', wig_path,
                '--budget', budget,
                '--gamma', gamma,
                '--tau', tau,
                '--c1', c1,
                '--c2', c2,
                '--c3', c3,
                '--solver', solver,
                '--bed_format', bed_format,
                '--outdir', outdir,
                '--rr_iter', N,
                '--identifiers', identifiers,
                '--verbose']

    if not verbose:
        cli_args.remove('--verbose')
    if identifiers is None:
        cli_args.remove('--identifiers')
        cli_args.remove(identifiers)
    return ' '.join(cli_args)

def main(args):
    chrom_args = get_params(args['param_file'], args['budget'],
                            args['gamma'], args['tau'], args['c1'],
                            args['c2'], args['c3'])

    args['outdir'] = rocco_aux.trim_path(args['outdir'])
    if not os.path.exists(args['outdir']):
        os.mkdir(args['outdir'])

    tmp = tempfile.NamedTemporaryFile(mode="w+")
    for i, arglist in enumerate(chrom_args):
        arglist = [str(x) for x in arglist]
        cmd = call_rocco(arglist[0], arglist[1], arglist[2], arglist[3],
                         arglist[4], arglist[5], arglist[6], arglist[7],
                         args['solver'], str(args['bed_format']),
                         args['verbose'], str(args['rr_iter']),
                         identifiers=args['identifiers'], outdir=args['outdir'])
        if not args['multi']:
            try:
                seq_process = subprocess.run(cmd.split(' '),
                                         capture_output=True, text=True, check=True)
                print(seq_process.stdout)
            except Exception as ex:
                print(ex)
                raise ex
        tmp.write(str(cmd + '\n'))

    tmp.flush()
    if args['multi']:
        rocco_aux.run_par(tmp.name, verbose=args['verbose'])
    tmp.close()

    if args['combine'] is not None:
        print('combining output files --> {}'.format(args['combine']))
        rocco_aux.sort_combine_bed(args['combine'], dir_=args['outdir'])


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--param_file',
                        required=True)
    parser.add_argument('-b', '--budget', type=float, default=.035, help='budget parameter (largest allowed fraction of selected bp) used for each chromosome with a `NULL` entry observed in `--param_file`')
    parser.add_argument('-g', '--gamma', type=float, default=1.0, help='gamma parameter (discontig. penalty weight) used for each chromosome  with a `NULL` entry observed in `--param_file`')
    parser.add_argument('-t', '--tau', type=float, default=0.0, help='tau parameter (enrichment threshold) used for each chromosome  with a `NULL` entry observed in `--param_file`')
    parser.add_argument('--c1', type=float, default=1.0, help='g_1 coefficient in score function (enrichment reward) used for each chromosome  with a `NULL` entry observed in `--param_file`')
    parser.add_argument('--c2', type=float, default=1.0, help='g_2 coefficient in score function (dispersion penalty) used for each chromosome  with a `NULL` entry observed in `--param_file`')
    parser.add_argument('--c3', type=float, default=1.0, help='g_3 coefficient in score function (local shift) used for each chromosome  with a `NULL` entry observed in `--param_file`')
    parser.add_argument('-N', '--rr_iter', type=int, default=50, help = 'number of RR iterations')
    parser.add_argument('--solver', default="ECOS",
                        help="Optimization software used to solve the \
                        main LP. `ECOS` is used by default.")
    parser.add_argument('--bed_format', type=int, default=3,
                        help="Specifies BED3 or BED6 format.\
                        Default is BED6. Generate BED3 output with \
                        --bed_format 3")
    parser.add_argument('--identifiers', default=None,
                        help="(optional) a filename containing identifiers\
                          for samples to include in experiment. Each identi\
                          fier should be a uniquely-identifying substring of\
                          the respective `.wig` sample. If not specified, all\
                          samples are used.")
    parser.add_argument('--outdir', default='.',
                        help="directory in which to store output bed files from the calls to rocco chrom")
    parser.add_argument('--combine', default=None, help="if not None, combine\
                        output bed files and store in the file specified\
                        with this parameter. ex: `--combine bedname.bed` con-\
                        catenates the chromosome-specific bedfiles into `bedname.bed`.")
    parser.add_argument('--multi', default=False, action='store_true', help='run rocco chrom jobs\
        simultaneously to improve speed. May increase peak memory use.')
    parser.add_argument('--verbose', default=False, action="store_true")
    args = vars(parser.parse_args())
    main(args)



"""
Obtain ROCCO conformable input from BAM files

This script is a ROCCO-specific wrapper for the PEPATAC (github.com/databio/pepatac)
tool `bamSitesToWig.py`. BAM files for each sample/replicate in `bamdir` are used to
create smooth, fixed-step bigwig signal tracks. These signal tracks are split by
chromosome, converted to human-readable .wig format, and then placed into directories
à la [this flowchart](https://github.com/nolan-h-hamilton/ROCCO/blob/main/docs/bamsig_flowchart.png).

The resulting `tracks_chr[]` directories can then be supplied to rocco chrom
via `--wig_path` to construct the signal matrix $\mathbf{S}_{chr}$.

```

Example:

    Run on toy data in `tests/data`:
    ```
    rocco prep --bamdir tests/data -s tests/data/test_sizes.sizes
    ```

"""

import os
import argparse
import subprocess
import pysam
from . import rocco_aux
import tempfile


def main(args):
    # process command line arguments
    cwd = os.getcwd()
    args['bamdir'] = rocco_aux.trim_path(os.path.abspath(args['bamdir']))
    args['outdir'] = rocco_aux.trim_path(os.path.abspath(args['outdir']))

    if not os.path.exists(args['sizes']):
        args['sizes'] = rocco_aux.get_size_file(args['sizes'])

    bigwig_files = []
    tf = tempfile.NamedTemporaryFile(mode='w', delete=False)
    for fname in os.listdir(args['bamdir']):
        fname = args['bamdir'] + '/' + fname
        fname = os.path.abspath(fname)
        if rocco_aux.is_alignment(fname):
            aln = pysam.AlignmentFile(fname)
            try:
                aln.check_index()
            except ValueError:
                print(f'no index file available for {fname}, calling pysam.index()')
                pysam.index(fname)
            print('{}: running bamSitesToWig.py'.format(aln.filename.decode()))
            bstw_path = os.path.join(os.path.dirname(__file__), "bamSitesToWig.py")
            bstw_cmd = ['python3', bstw_path,
                        '-i', aln.filename.decode(),
                        '-c', args['sizes'],
                        '-w', args['outdir'] + '/' + aln.filename.decode().split('/')[-1] + '.bw',
                        '-r', str(args['interval_length']),
                        '-m', 'atac',
                        '-p', str(args['cores']),
                        '--variable-step']
            if not args['multi']:
                # if `--multi False
                subprocess.run(bstw_cmd,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL,
                               shell=False, check=True)
            else:
                tf.write(' '.join(bstw_cmd))
                tf.write('\n')

            bigwig_file = args['outdir'] + '/' + aln.filename.decode().split('/')[-1] + '.bw'
            bigwig_files.append(bigwig_file)
    tf.close()
    if args['multi']:
        rocco_aux.run_par(tf.name)

    chroms = list(rocco_aux.parse_size_file(args['sizes']).keys())
    os.chdir(args['outdir'])
    for chrom in chroms:
        try:
            os.mkdir(f'tracks_{chrom}')
        except FileExistsError:
            pass
        for bigwig_file in [bw_file.split('/')[-1] for bw_file in bigwig_files]:
            subprocess.run(['bigWigToWig', bigwig_file,
                        chrom + '_' + bigwig_file + '.wig',
                        '-chrom={}'.format(chrom)], check=True)
            subprocess.run(["mv", chrom + '_' + bigwig_file + '.wig',
                        "tracks_{}".format(chrom)], check=True)
    os.chdir(cwd)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='BAM preprocessing')
    parser.add_argument('-i','--bamdir', default='.', type=str, help='path to directory containing BAM files')
    parser.add_argument('-o', '--outdir', default= '.', help='output directory')
    parser.add_argument('-s', '--sizes',
                        default='hg38',
                        help="A path to a chromosome sizes file. OR\
                        an assembly name")
    parser.add_argument('-L', '--interval_length',
                        default=50,
                        help="wiggle track fixed interval length")
    parser.add_argument('-c', '--cores',
                        type=int,
                        default=1,
                        help="`bamSitesToWig.py`'s cores  parameter.\
                        Altering this parameter to use >1 core\
                        may cause issues on Mac OS")
    parser.add_argument('--multi',
                         default=True,
                         help='Set to `False` to run `bamSitesToWig` jobs\
                         sequentially.')
    parser.add_argument('--index',
                        type=int,
                        default=1,
                        help="Deprecated -- backwards compatibility")
    parser.add_argument('--bstw_path', default=None, help="Deprecated -- backwards compatibility")
    args = vars(parser.parse_args())
    main(args)



"""
Set budgets for each chromosome based on average read density observed
across multiple BAM files in directory `--bamdir`. Output (stdout) is formatted
for easy use as a parameter CSV file for `rocco gwide` (`--param_file`). Note: the
output CSV assumes wig track directories are local. Adjust paths accordingly.


Examples:
    ```
    rocco budgets -i ../bamfiles -s hg38.sizes -a 0 -b .05
        chromosome,input_path,budget,gamma,tau,c1,c2,c3
        chr1,tracks_chr1,0.037,NULL,NULL,NULL,NULL,NULL
        chr2,tracks_chr2,0.038,NULL,NULL,NULL,NULL,NULL
        chr3,tracks_chr3,0.039,NULL,NULL,NULL,NULL,NULL
        chr4,tracks_chr4,0.036,NULL,NULL,NULL,NULL,NULL
        chr5,tracks_chr5,0.037,NULL,NULL,NULL,NULL,NULL
        chr6,tracks_chr6,0.039,NULL,NULL,NULL,NULL,NULL
        chr7,tracks_chr7,0.037,NULL,NULL,NULL,NULL,NULL
        chr8,tracks_chr8,0.038,NULL,NULL,NULL,NULL,NULL
        chr9,tracks_chr9,0.032,NULL,NULL,NULL,NULL,NULL
        chr10,tracks_chr10,0.038,NULL,NULL,NULL,NULL,NULL
        chr11,tracks_chr11,0.041,NULL,NULL,NULL,NULL,NULL
        chr12,tracks_chr12,0.04,NULL,NULL,NULL,NULL,NULL
        chr13,tracks_chr13,0.03,NULL,NULL,NULL,NULL,NULL
        chr14,tracks_chr14,0.032,NULL,NULL,NULL,NULL,NULL
        chr15,tracks_chr15,0.029,NULL,NULL,NULL,NULL,NULL
        chr16,tracks_chr16,0.036,NULL,NULL,NULL,NULL,NULL
        chr17,tracks_chr17,0.043,NULL,NULL,NULL,NULL,NULL
        chr18,tracks_chr18,0.035,NULL,NULL,NULL,NULL,NULL
        chr19,tracks_chr19,0.05,NULL,NULL,NULL,NULL,NULL
        chr20,tracks_chr20,0.04,NULL,NULL,NULL,NULL,NULL
        chr21,tracks_chr21,0.028,NULL,NULL,NULL,NULL,NULL
        chr22,tracks_chr22,0.031,NULL,NULL,NULL,NULL,NULL
        chrX,tracks_chrX,0.023,NULL,NULL,NULL,NULL,NULL
        chrY,tracks_chrY,0.0,NULL,NULL,NULL,NULL,NULL
    ```

    ```
    rocco budgets -i ../bamfiles -s hg38.sizes --desired_avg .035
        chromosome,input_path,budget,gamma,tau,c1,c2,c3
        chr1,tracks_chr1,0.037,NULL,NULL,NULL,NULL,NULL
        chr2,tracks_chr2,0.039,NULL,NULL,NULL,NULL,NULL
        chr3,tracks_chr3,0.039,NULL,NULL,NULL,NULL,NULL
        chr4,tracks_chr4,0.036,NULL,NULL,NULL,NULL,NULL
        chr5,tracks_chr5,0.037,NULL,NULL,NULL,NULL,NULL
        chr6,tracks_chr6,0.039,NULL,NULL,NULL,NULL,NULL
        chr7,tracks_chr7,0.038,NULL,NULL,NULL,NULL,NULL
        chr8,tracks_chr8,0.038,NULL,NULL,NULL,NULL,NULL
        chr9,tracks_chr9,0.033,NULL,NULL,NULL,NULL,NULL
        chr10,tracks_chr10,0.039,NULL,NULL,NULL,NULL,NULL
        chr11,tracks_chr11,0.041,NULL,NULL,NULL,NULL,NULL
        chr12,tracks_chr12,0.04,NULL,NULL,NULL,NULL,NULL
        chr13,tracks_chr13,0.031,NULL,NULL,NULL,NULL,NULL
        chr14,tracks_chr14,0.033,NULL,NULL,NULL,NULL,NULL
        chr15,tracks_chr15,0.03,NULL,NULL,NULL,NULL,NULL
        chr16,tracks_chr16,0.037,NULL,NULL,NULL,NULL,NULL
        chr17,tracks_chr17,0.043,NULL,NULL,NULL,NULL,NULL
        chr18,tracks_chr18,0.035,NULL,NULL,NULL,NULL,NULL
        chr19,tracks_chr19,0.05,NULL,NULL,NULL,NULL,NULL
        chr20,tracks_chr20,0.04,NULL,NULL,NULL,NULL,NULL
        chr21,tracks_chr21,0.029,NULL,NULL,NULL,NULL,NULL
        chr22,tracks_chr22,0.032,NULL,NULL,NULL,NULL,NULL
        chrX,tracks_chrX,0.024,NULL,NULL,NULL,NULL,NULL
        chrY,tracks_chrY,0.002,NULL,NULL,NULL,NULL,NULL
    ```
"""

import argparse
import numpy as np
import os
from typing import Dict
import pysam
from . import rocco_aux


def rd_dens(bamfile: str, size_file: str, a: float = 0.0, b: float = 0.05, desired_avg: float = -1.0, index: bool = False) -> Dict[str, float]:
    """
    Calculate chromosome-specific budgets as scaled read density for `bamfile`

    Args:
        bamfile (str): Path to the BAM file.
        rdlen (int): Read length
        size_file (str): path to a chromosome sizes file
        a (float): Minimum allowed budget for a particular chromosome
            this bound is ignored if `desired_avg` is non-negative.
        b (float): Maximum allowed budget for a particular chromosome
            this bound is ignored if `desired_avg` is non-negative.
        desired_avg (float): scaled read densities will average to this value
            if it is modified to be non-negative.
        index (bool): whether to index `bamfile` with pysam.index()

    Returns:
        dict: A dictionary {chrom: budget}
    """
    if not os.path.exists(size_file):
        raise FileNotFoundError(f'Could not find size file {size_file}')

    size_dict = rocco_aux.parse_size_file(size_file)
    if index:
        pysam.index(bamfile)

    aln = pysam.AlignmentFile(bamfile,'rb')
    chr_reads = {}
    try:
        for chrom_record in aln.get_index_statistics():
            if chrom_record[0] not in size_dict.keys():
                continue
            chr_reads.update({chrom_record[0]: chrom_record[1]})
    except ValueError as no_index_err:
        print(f'\nest_budgets.rd_dens: could not find index file for {bamfile}. Run script with --index argument\n')
        raise no_index_err

    for key in chr_reads.keys():
        chr_reads[key] /= size_dict[key]

    # if `desired_avg` is nonnegative, scale read densities such that their mean is `desired_avg``
    current_avg = np.mean(list(chr_reads.values()))
    if desired_avg >= 0:
        for key in chr_reads.keys():
            chr_reads[key] *= desired_avg/current_avg
        return chr_reads

    # (default) apply min-max normalization to scale read densities in [a,b] if desired_avg < 0 
    min_val = min(chr_reads.values())
    max_val = max(chr_reads.values())
    for key in chr_reads.keys():
        chr_reads[key] = (((chr_reads[key] - min_val) / (max_val - min_val)) * (b - a) + a)
    return chr_reads


def avg_rd(bamdir: str, size_file: str, a: float = 0.0, b: float = 0.05, desired_avg: float = -1, index: bool = False) -> Dict[str, float]:
    """
    Calculate the average read density across *multiple* BAM files in `bamdir`.

    Args:
        bamdir (str): Path to the directory containing BAM files.
        size_file (str): path to a chromosome sizes file
        a (float): Minimum allowed budget for a particular chromosome
            this bound is ignored if `desired_avg` is non-negative.
        b (float): Maximum allowed budget for a particular chromosome
            this bound is ignored if `desired_avg` is non-negative.
        desired_avg (float): scaled read densities will average to this value
            if it is modified to be non-negative.

    Returns:
        dict: A dictionary {chrom: budget}
    """
    bamdir = rocco_aux.trim_path(bamdir)
    bamfiles = [f'{bamdir}/{x}' for x in os.listdir(bamdir)
                if x.split('.')[-1] == 'bam']
    bam_rd_dicts = [rd_dens(x, size_file=size_file, a=a, b=b, desired_avg=desired_avg, index=index) for x in bamfiles]
    final_dict = dict.fromkeys(bam_rd_dicts[0].keys(), 0)

    for key in final_dict.keys():
        for bam_rd_dict in bam_rd_dicts:
            final_dict[key] += bam_rd_dict[key] / len(bam_rd_dicts)
    return final_dict


def main(args):

    dict_output = avg_rd(args['bamdir'], args['sizes'],  a=args['a'], b=args['b'], desired_avg=args['desired_avg'], index=args['index'])
    print('chromosome,input_path,budget,gamma,tau,c1,c2,c3')
    for key, val in dict_output.items():
        print(f'{key},tracks_{key},{round(val, 3)},NULL,NULL,NULL,NULL,NULL')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Compute chromosome-specific budget parameters based on observed read densities.\
        Uses min-max normalization (default) on the read density vals to yield budgets in interval [a,b] OR scales the values by a\
        constant such that their mean is `desired_avg`.')
    parser.add_argument('-i', '--bamdir', type=str, help='Path to the directory containing .bam and .bai files')
    parser.add_argument('-s', '--sizes', type=str, help='chromosome sizes file')
    parser.add_argument('-a', type=float, default=0.0, help='Minimum allowed budget. This bound is ignored if\
        `--desired_avg` is non-negative.')
    parser.add_argument('-b', type=float, default=0.05, help='Maximum allowed budget. This bound is ignored if\
        `--desired_avg` is non-negative.')
    parser.add_argument('--desired_avg', type=float, default=-1.0, help='Scaled read densities (i.e., budgets) will\
        average to this value if non-negative. Defaults to -1.')
    parser.add_argument('--index', default=False, action='store_true', help='invoke `--index` if BAM files are not yet indexed')
    args = vars(parser.parse_args())

    if args['sizes'] is None:
        print('rocco budgets: A .sizes file is required for the corresponding genome.')
        print('\tUse the `get_sizes` subcommand or retrieve a sizes file manually')
        raise Exception('a sizes file is required. Consider `rocco get_sizes -g [assembly]`')

    main(args)



"""
Run ROCCO on singe particular chromosome `--chrom`.
This script is the workhorse for `rocco gwide` which generates and compiles
results for multiple chromosomes
"""

import os
import gc
import argparse
import warnings
import subprocess
import numpy as np
import pandas as pd
from .locus import Locus
from .loci import Loci



def get_locus_size(wig_path) -> int:
    """
    Infers step size from replicates` wig files.

    Args:
        wig_path (str) : path to *directory* containing wig files


    Returns:
        int: an integer locus/step size used in the wig files

    """
    i = 0
    pos1 = 0
    pos2 = 0
    wig_file = [x for x in os.listdir(wig_path) if '.wig' in x][0]
    with open(wig_path + '/' + wig_file, mode='r', encoding='utf-8') as wig:
        for line in wig:
            line = line.strip()
            if line.replace('\t', '').replace(' ', '').isnumeric():
                if i == 0:
                    pos1 = int(line.split('\t')[0])
                if i == 1:
                    pos2 = int(line.split('\t')[0])
                if pos2 - pos1 > 0:
                    break
                i += 1
    return pos2 - pos1

def get_start_end(wig_path):
    """
    Infers starting and ending nucleotide position common to the
    multiple signal tracks

    To infer a reasonable genomic region over which to run ROCCO,
    this function returns the median starting point and endpoints
    among the multiple replicates' signal files in `wig_path`.

    Args:
        wig_path (str) : path to *directory* containing wig files

    Returns:
        start (int): median starting position
        end (int): median ending position
    """
    starts = []
    ends = []

    def get_start(track):
        with open(track, mode='r', encoding='utf-8') as wig:
            for line in wig:
                line = line.strip()

                if line.replace('\t', '').replace(' ', '').isnumeric():
                    start = int(line.split('\t')[0])
                    return start
        return None

    def get_end(track):
        with subprocess.Popen(["tail", "-n1", track],
                              stdout=subprocess.PIPE) as proc:
            for line in proc.stdout:
                line = line.decode()
                line = line.strip()
                return int(line.split('\t')[0])
        return None

    wig_files = [x for x in os.listdir(wig_path) if '.wig' in x]
    for wig_file in wig_files:
        starts.append(get_start(wig_path + '/' + wig_file))
        ends.append(get_end(wig_path + '/' + wig_file))
    return sorted(starts)[len(starts) // 2], sorted(ends)[len(ends) // 2]


def read_wig(wig_file, start=0, end=10**10, locus_size=50):
    r"""
    Processes a wig file for inclusion in the signal matrix $\mathbf{S}_{chr}$

    This function prepares signal data for to fit in a constant-sized
    $K \times n$ matrix. If a wig file begins at a position < `start`,
    it will be padded with zeros. Likewise if a wig file ends at a
    position > `end`, the extra values are ignored. Gaps between `start`
    and `end` are replaced with a zero-entry.

    Args:
        wig_file (str): path to wiggle formatted signal track
        start (int): inferred or manually-specified starting nucleotide
          position
        end (int): inferred or manually-specified ending nucleotide
          position
        locus_size (int): interval/locus size. Each wig file will have
          signal values at every `start + i*locus_size` nucleotide pos.
          for i = 0,1,2,...

    Returns:
        loci: list of starting nucleotide positions for each locus
        signal: enrichment signal value at each locus in `loci`
    """
    log("ROCCO_chrom: reading wig file {}".format(wig_file))
    loci = []
    signal = []
    with open(wig_file, mode='r', encoding='utf-8') as wig:
        for line in wig:
            line = line.strip()
            try:
                line = line.split('\t')
                line[0] = int(line[0])
                line[1] = float(line[1])
                if line[1] < 0:
                    line[1] = 0
                # case: wig begins after specified `start`
                if start < line[0] and len(loci) == 0:
                    for loc in range(start, line[0], locus_size):
                        loci.append(loc)
                        signal.append(0)
                    loci.append(line[0])
                    signal.append(line[1])
                    continue
                if line[0] < start:
                    continue
                if line[0] > end:
                    break
                # case: loci not contiguous
                if len(loci) > 0 and line[0] - loci[-1] > locus_size:
                    loci.append(loci[-1] + locus_size)
                    signal.append(0)
                    continue

                loci.append(line[0])
                signal.append(line[1])

            except ValueError:
                continue
            except IndexError:
                continue
    # case: wig ends prematurely
    if loci[-1] < end:
        for loc in range(loci[-1] + locus_size, end + locus_size, locus_size):
            loci.append(loc)
            signal.append(0)

    return loci, signal


def collect_wigs(wig_files, start=0, end=10**10, locus_size=50) -> pd.DataFrame:
    r"""
    creates a $K \times n$ dataframe of signals derived from wig files

    Args:
        wig_files (list): list of wig filepaths
        start (int): starting nucleotide position of wig files
        end (int): ending nucleotide position of wig files
        locus_size (int): initial size (in nucleotides) of Locus objects.
            Corresponds to $L$ in paper.

    Returns:
        pd.DataFrame: $K \times n$ dataframe of individuals' signals.
    """
    loci = [x for x in range(start, end + locus_size, locus_size)]
    all_wigs = pd.DataFrame(
        np.array([read_wig(x, start, end, locus_size)[1]
                  for x in wig_files],dtype=np.float32),
        columns=loci, index=None)
    return all_wigs


def emp_cdf(val, arr) -> float:
    """
    Get empirical cdf of `val` in `arr`

    Used to compute 0-1000 scaled score for BED6 files
    and visualization in genome browser.

    Note:
        Assumes `arr` is sorted!

    Args:
        val (float): a numeric element in `arr`
        arr (array): a SORTED array of peak values

    Returns:
        (float) estimate for percentile of `val` in `arr`
    """
    return np.searchsorted(arr, val) / len(arr)


def sig_selected(loci_obj: Loci) -> np.ndarray:
    """
    Get enrichment signals from all selected loci

    Args:
        loci_obj (Loci): Loci object after optimization

    Returns:
        np.ndarray: vector of enrichment signals from all selected
            (but still unmerged) loci
    """
    sig_ref = []
    for loc in loci_obj:
        if loc.accessible == 1:
            sig_ref.append(np.sum(loc.sig_data) / loc.size)
    return np.array(sig_ref)


def tags(fname) -> list:
    """
    Get sample IDs from file specified with command-line argument `--identifiers`

    Args:
        fname (str): path to file containing a subset of sample IDs--one line for each ID

    Returns:
        list: list of sample IDs in `fname`
    """
    tags_ = []
    for line in open(fname, mode='r', encoding='utf-8'):
        tags_.append(line.strip())
    return tags_


def log(text, verbose=True) -> None:
    """
    Only print `text` if `verbose=True`.
    """
    if verbose:
        print(str(text))


def main(args):
    log('ROCCO_chrom: args', args['verbose'])
    log(args, args['verbose'])

    if args['locus_size'] == -1:
        args['locus_size'] = get_locus_size(args['wig_path'])

    if args['outdir'][-1] == '/':
        args['outdir'] = args['outdir'][0:-1]

    if args['start'] == -1 or args['end'] == -1:
        args['start'], args['end'] = get_start_end(args['wig_path'])
    if args['bed_format'] not in [3, 6]:
        warnings.warn('Only BED3 and BED6 are supported--setting to BED3')
        args['bed_format'] = 6

    log("ROCCO_chrom: inferred args", args['verbose'])
    log(args, args['verbose'])

    # collect wig files for each replicate/sample for `--chrom`
    wig_files = [args['wig_path'] + '/' + fname
                 for fname in os.listdir(args['wig_path'])
                 if 'wig' in fname.split('.')[-1]]

    # select a subset of the wig files if `--identifiers` is defined
    if args['identifiers'] is not None:
        new_wig_files = []
        for tag in tags(args['identifiers']):
            for wfname in wig_files:
                if tag in wfname:
                    new_wig_files.append(wfname)
        wig_files = new_wig_files

    # create a dataframe of signal values for `K` replicates and `n` loci
    signal_matrix = collect_wigs(wig_files,
                                 start=args['start'],
                                 end=args['end'],
                                 locus_size=args['locus_size'])

    log("ROCCO_chrom: building Loci object", args['verbose'])
    InitLoci = Loci()
    for i, loc in enumerate(signal_matrix.columns):
        size_ = args['locus_size']
        NewLoc = Locus(position=args['start'] + i * size_,
                       size=size_,
                       sig_data=np.array(signal_matrix[loc],dtype=np.float32))
        InitLoci.append_locus(NewLoc)
    del signal_matrix
    gc.collect()

    log("ROCCO_chrom: solving for optimal solution", args['verbose'])
    InitLoci.rocco_lp(budget=args['budget'],
                        gam=args['gamma'],
                        tau=args['tau'],
                        c1=args['c1'],
                        c2=args['c2'],
                        c3=args['c3'],
                        verbose_=args['verbose'],
                        solver=args['solver'],
                        N=args['rr_iter'])

    log("ROCCO_chrom: merging adjacent selections", args['verbose'])
    InitLoci.combine_selected()

    if not os.path.exists(args['outdir']):
        os.mkdir(args['outdir'])

    fname = args['outdir'] + '/' + "ROCCO_out_{}_{}_{}_{}_{}_{}_{}.bed".format(
        args['chrom'],
        args['budget'],
        args['gamma'],
        args['tau'],
        args['c1'],
        args['c2'],
        args['c3'])

    log('ROCCO_chrom: writing output: {}'.format(fname), True)
    outfile = open(fname, mode='w', encoding='utf-8')
    output = ""
    if args['bed_format'] == 6:
        loc_count = 1
        sorted_sel_sig = np.sort(sig_selected(InitLoci))
        for loc in InitLoci:
            assert loc.size > 0
            if loc.accessible == 1:
                # the score can be used for coloring in ucsc genome browser
                loc_score = int(500 * emp_cdf(np.sum(loc.sig_data) / loc.size,
                                              sorted_sel_sig) + 500)
                loc_name = args['chrom'] + '_' + str(loc_count)
                output += "{}\t{}\t{}\t{}\t{}\t{}\n".format(
                    args['chrom'], loc.position, loc.position + loc.size,
                    loc_name, loc_score, '.')
                loc_count += 1

    if args['bed_format'] == 3:
        for loc in InitLoci:
            if loc.accessible == 1:
                output += "{}\t{}\t{}\n".format(
                    args['chrom'], loc.position, loc.position + loc.size)

    outfile.write(output)
    outfile.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--chrom', type=str, help="e.g., --chrom chr1")
    parser.add_argument('--start', type=int, default=-1, help='beginning nucleotide position (default: infer from wiggle files)')
    parser.add_argument('--end', type=int, default=-1, help='ending nucleotide position (default: infer from wiggle files)')
    parser.add_argument('--locus_size', type=int, default=-1, help='size of loci (L parameter) (default: infer from wiggle files)')
    parser.add_argument('--wig_path', type=str, default=os.getcwd(), help='directory containing .wig files for each sample')
    parser.add_argument('-N', '--rr_iter', type=int, default=50, help = 'number of RR iterations')
    parser.add_argument('--verbose', default=False, action="store_true", help='set to `True` for verbose logging')
    parser.add_argument('-b', '--budget', type=float, default=.035, help='budget parameter (largest allowed fraction of selected bp)')
    parser.add_argument('-g', '--gamma', type=float, default=1.0, help='gamma parameter (discontig. penalty weight)')
    parser.add_argument('-t', '--tau', type=float, default=0.0, help='tau parameter (enrichment threshold)')
    parser.add_argument('--c1', type=float, default=1.0, help='g_1 coefficient in score function (enrichment reward)')
    parser.add_argument('--c2', type=float, default=1.0, help='g_2 coefficient in score function (dispersion penalty) ')
    parser.add_argument('--c3', type=float, default=1.0, help='g_3 coefficient in score function (local shift)')
    parser.add_argument('--solver', default="ECOS", help='Optimization software used to solve the \
                        main LP. `ECOS` is used by default.')
    parser.add_argument('--bed_format', type=int, default=6,
                        help="`3` for BED3 format and `6` for BED6 format")
    parser.add_argument('--identifiers', default=None,
                        help="(optional) a filename containing identifiers\
                          for samples to include in experiment. Each identi\
                          fier should be a uniquely-identifying substring of\
                          the respective `.wig` sample. If not specified, all\
                          samples are used.")
    parser.add_argument('--outdir', type=str, default='.', help="directory in which to store output bed file")

    args = vars(parser.parse_args())
    main(args)



import linecache
import os
import sys
import shlex
import subprocess
from pybedtools import BedTool
def wrap_run(cmd):
    print(cmd)
    try:
        cmd_list = shlex.split(cmd)
        proc = subprocess.run(cmd_list,
            shell=False,
            capture_output=True,
            text=True)
        return proc
    except Exception as e:
        print(f"Error executing command: {e}")
        return None

def line_k(fname, k):
    try:
        line = linecache.getline(fname, k)
        line = line.strip()
        return line
    except FileNotFoundError:
        return None

def is_nonempty(fname, min_bytes=10):
    return os.path.getsize(fname) >= min_bytes

def log(txt, script):
    print(f'{script}: {txt}')


wig_min_bytes = 100
min_wig_line_ct = 100
jac_min = .99

log('ensuring cwd is `tests` and parent directory is `ROCCO`...',
    sys.argv[0])
assert os.path.abspath(os.getcwd())[-5:] == 'tests'
assert os.path.abspath(os.path.pardir)[-5:] == 'ROCCO'

log('changing current working directory to parent dir.',
    sys.argv[0])
os.chdir(os.path.pardir)
assert os.path.abspath(os.getcwd())[-5:] == 'ROCCO'

log('running ROCCO_chrom.py for chr21, store output in `tests/output`', sys.argv[0])
wrap_run('rocco chrom --chrom chr21 --wig_path tests/data/tracks_chr21 -b .02 -N 0 --outdir tests/output')
log('running ROCCO_chrom.py for chr22, store output in `tests/output`', sys.argv[0])
wrap_run('rocco chrom --chrom chr22 --wig_path tests/data/tracks_chr22 -b .02 -N 0 --outdir tests/output')

assert os.path.exists('tests/output/ROCCO_out_chr21_0.02_1.0_0.0_1.0_1.0_1.0.bed')
chr21_bed = BedTool('tests/output/ROCCO_out_chr21_0.02_1.0_0.0_1.0_1.0_1.0.bed')
assert os.path.exists('tests/output/ROCCO_out_chr22_0.02_1.0_0.0_1.0_1.0_1.0.bed')
chr22_bed = BedTool('tests/output/ROCCO_out_chr22_0.02_1.0_0.0_1.0_1.0_1.0.bed')
merged_bed = chr21_bed.cat(chr22_bed)

log('running ROCCO_gwide.py using tests/test_params.csv, store output in `tests/output` and combine', sys.argv[0])
wrap_run('rocco gwide -p tests/test_params.csv --outdir tests/output/combined --combine tests/combined.bed -N 0 --multi')
assert os.path.exists('tests/combined.bed')
combined_bed = BedTool('tests/combined.bed')

log('Comparing output from the ROCCO_chrom.py jobs and ROCCO.py job to ensure similarity', sys.argv[0])
assert merged_bed.jaccard(combined_bed)['jaccard'] >= jac_min

log('all cases passed', sys.argv[0])